# -*- coding: utf-8 -*-
"""
Created on Wed Jun 28 18:05:45 2023

@author: richie bao

ref NLP FROM SCRATCH: TRANSLATION WITH A SEQUENCE TO SEQUENCE NETWORK AND ATTENTION, https://pytorch.org/tutorials/intermediate/seq2seq_translation_tutorial.html
"""
from __future__ import unicode_literals, print_function, division
from io import open
import unicodedata
import re
import random
import os

import torch
import torch.nn as nn
from torch import optim
import torch.nn.functional as F

import numpy as np
from torch.utils.data import TensorDataset, DataLoader, RandomSampler

import matplotlib.pyplot as plt
# plt.switch_backend('agg')
import matplotlib.ticker as ticker
import matplotlib
import numpy as np

import time
import math

if __package__:
    from ._nlp_tools import prepareData4seq2seq  
else:
    from _nlp_tools import prepareData4seq2seq
        
SOS_token = 0
EOS_token = 1

class EncoderRNN(nn.Module):
    def __init__(self, input_size, hidden_size, dropout_p=0.1):
        super(EncoderRNN, self).__init__()
        self.hidden_size = hidden_size

        self.embedding = nn.Embedding(input_size, hidden_size)
        self.gru = nn.GRU(hidden_size, hidden_size, batch_first=True)
        self.dropout = nn.Dropout(dropout_p)

    def forward(self, input):
        embedded = self.dropout(self.embedding(input))
        output, hidden = self.gru(embedded)
        return output, hidden
    
class DecoderRNN(nn.Module):
    def __init__(self, hidden_size, output_size):
        super(DecoderRNN, self).__init__()
        self.embedding = nn.Embedding(output_size, hidden_size)
        self.gru = nn.GRU(hidden_size, hidden_size, batch_first=True)
        self.out = nn.Linear(hidden_size, output_size)

    def forward(self, encoder_outputs, encoder_hidden, target_tensor=None):
        batch_size = encoder_outputs.size(0)
        decoder_input = torch.empty(batch_size, 1, dtype=torch.long, device=device).fill_(SOS_token)
        decoder_hidden = encoder_hidden
        decoder_outputs = []

        for i in range(MAX_LENGTH):
            decoder_output, decoder_hidden  = self.forward_step(decoder_input, decoder_hidden)
            decoder_outputs.append(decoder_output)

            if target_tensor is not None:
                # Teacher forcing: Feed the target as the next input
                decoder_input = target_tensor[:, i].unsqueeze(1) # Teacher forcing
            else:
                # Without teacher forcing: use its own predictions as the next input
                _, topi = decoder_output.topk(1)
                decoder_input = topi.squeeze(-1).detach()  # detach from history as input

        decoder_outputs = torch.cat(decoder_outputs, dim=1)
        decoder_outputs = F.log_softmax(decoder_outputs, dim=-1)
        return decoder_outputs, decoder_hidden, None # We return `None` for consistency in the training loop

    def forward_step(self, input, hidden):
        output = self.embedding(input)
        output = F.relu(output)
        output, hidden = self.gru(output, hidden)
        output = self.out(output)
        return output, hidden    

class BahdanauAttention(nn.Module):
    def __init__(self, hidden_size):
        super(BahdanauAttention, self).__init__()
        self.Wa = nn.Linear(hidden_size, hidden_size)
        self.Ua = nn.Linear(hidden_size, hidden_size)
        self.Va = nn.Linear(hidden_size, 1)

    def forward(self, query, keys):
        scores = self.Va(torch.tanh(self.Wa(query) + self.Ua(keys)))
        scores = scores.squeeze(2).unsqueeze(1)

        weights = F.softmax(scores, dim=-1)
        context = torch.bmm(weights, keys)

        return context, weights

class AttnDecoderRNN(nn.Module):
    def __init__(self, hidden_size, output_size, dropout_p=0.1,MAX_LENGTH=10,device='cpu'): # device='cpu'
        super(AttnDecoderRNN, self).__init__()
        self.MAX_LENGTH=MAX_LENGTH
        self.device=device
        self.embedding = nn.Embedding(output_size, hidden_size)
        self.attention = BahdanauAttention(hidden_size)
        self.gru = nn.GRU(2 * hidden_size, hidden_size, batch_first=True)
        self.out = nn.Linear(hidden_size, output_size)
        self.dropout = nn.Dropout(dropout_p)

    def forward(self, encoder_outputs, encoder_hidden, target_tensor=None):
        batch_size = encoder_outputs.size(0)
        decoder_input = torch.empty(batch_size, 1, dtype=torch.long,device=self.device).fill_(SOS_token) 
        decoder_hidden = encoder_hidden
        decoder_outputs = []
        attentions = []

        for i in range(self.MAX_LENGTH):
            decoder_output, decoder_hidden, attn_weights = self.forward_step(
                decoder_input, decoder_hidden, encoder_outputs
            )
            decoder_outputs.append(decoder_output)
            attentions.append(attn_weights)

            if target_tensor is not None:
                # Teacher forcing: Feed the target as the next input
                decoder_input = target_tensor[:, i].unsqueeze(1) # Teacher forcing
            else:
                # Without teacher forcing: use its own predictions as the next input
                _, topi = decoder_output.topk(1)
                decoder_input = topi.squeeze(-1).detach()  # detach from history as input

        decoder_outputs = torch.cat(decoder_outputs, dim=1)
        decoder_outputs = F.log_softmax(decoder_outputs, dim=-1)
        attentions = torch.cat(attentions, dim=1)

        return decoder_outputs, decoder_hidden, attentions


    def forward_step(self, input, hidden, encoder_outputs):
        embedded =  self.dropout(self.embedding(input))

        query = hidden.permute(1, 0, 2)
        context, attn_weights = self.attention(query, encoder_outputs)
        input_gru = torch.cat((embedded, context), dim=2)

        output, hidden = self.gru(input_gru, hidden)
        output = self.out(output)

        return output, hidden, attn_weights
    
def indexesFromSentence(lang, sentence):
    return [lang.word2index[word] for word in sentence.split(' ')]

def tensorFromSentence(lang, sentence,device):
    indexes = indexesFromSentence(lang, sentence)
    indexes.append(EOS_token)
    return torch.tensor(indexes, dtype=torch.long, device=device).view(1, -1)

def tensorsFromPair(pair,device):
    input_tensor = tensorFromSentence(input_lang, pair[0],device)
    target_tensor = tensorFromSentence(output_lang, pair[1],device)
    return (input_tensor, target_tensor)

def get_dataloader4seq2seq(batch_size,root,lang1, lang2,eng_prefixes,MAX_LENGTH=10,device='cpu',reverse=False,cmn=False):
    input_lang, output_lang, pairs = prepareData4seq2seq(root,lang1, lang2, eng_prefixes,MAX_LENGTH,reverse,cmn)

    n = len(pairs)
    input_ids = np.zeros((n, MAX_LENGTH), dtype=np.int32)
    target_ids = np.zeros((n, MAX_LENGTH), dtype=np.int32)

    for idx, (inp, tgt) in enumerate(pairs):
        inp_ids = indexesFromSentence(input_lang, inp)
        tgt_ids = indexesFromSentence(output_lang, tgt)
        inp_ids.append(EOS_token)
        tgt_ids.append(EOS_token)
        input_ids[idx, :len(inp_ids)] = inp_ids
        target_ids[idx, :len(tgt_ids)] = tgt_ids

    train_data = TensorDataset(torch.LongTensor(input_ids).to(device),
                               torch.LongTensor(target_ids).to(device))

    train_sampler = RandomSampler(train_data)
    train_dataloader = DataLoader(train_data, sampler=train_sampler, batch_size=batch_size)
    return input_lang, output_lang, pairs,train_dataloader    

def train_epoch(dataloader, encoder, decoder, encoder_optimizer,
          decoder_optimizer, criterion):

    total_loss = 0
    for data in dataloader:
        input_tensor, target_tensor = data

        encoder_optimizer.zero_grad()
        decoder_optimizer.zero_grad()

        encoder_outputs, encoder_hidden = encoder(input_tensor)
        decoder_outputs, _, _ = decoder(encoder_outputs, encoder_hidden, target_tensor)

        loss = criterion(
            decoder_outputs.view(-1, decoder_outputs.size(-1)),
            target_tensor.view(-1)
        )
        loss.backward()

        encoder_optimizer.step()
        decoder_optimizer.step()

        total_loss += loss.item()

    return total_loss / len(dataloader)

def asMinutes(s):
    m = math.floor(s / 60)
    s -= m * 60
    return '%dm %ds' % (m, s)

def timeSince(since, percent):
    now = time.time()
    s = now - since
    es = s / (percent)
    rs = es - s
    return '%s (- %s)' % (asMinutes(s), asMinutes(rs))

def seq2seq_train(train_dataloader, encoder, decoder, n_epochs, learning_rate=0.001,
               print_every=100, plot_every=100,save_path=None):    
    start = time.time()
    plot_losses = []
    print_loss_total = 0  # Reset every print_every
    plot_loss_total = 0  # Reset every plot_every

    encoder_optimizer = optim.Adam(encoder.parameters(), lr=learning_rate)
    decoder_optimizer = optim.Adam(decoder.parameters(), lr=learning_rate)
    criterion = nn.NLLLoss()    
    
    if os.path.exists(save_path):
        checkpoint= torch.load(save_path)
        encoder.load_state_dict(checkpoint['best_encoder_state_dict'])
        decoder.load_state_dict(checkpoint['best_decoder_state_dict'])
        encoder_optimizer.load_state_dict(checkpoint['encoder_optimizer_state_dict'])
        decoder_optimizer.load_state_dict(checkpoint['decoder_optimizer_state_dict'])
        epoch_trained = checkpoint['epoch']
        print('loaded checkpoint')
    else:
        epoch_trained=0          

    best_loss=np.inf
    for epoch in range(1, n_epochs + 1):
        loss = train_epoch(train_dataloader, encoder, decoder, encoder_optimizer, decoder_optimizer, criterion)
        print_loss_total += loss
        plot_loss_total += loss

        if epoch % print_every == 0:
            print_loss_avg = print_loss_total / print_every
            print_loss_total = 0
            print('%s (%d %d%%) %.4f' % (timeSince(start, epoch / n_epochs),
                                        epoch, epoch / n_epochs * 100, print_loss_avg))

        if epoch % plot_every == 0:
            plot_loss_avg = plot_loss_total / plot_every
            plot_losses.append(plot_loss_avg)
            plot_loss_total = 0
            
        if loss<best_loss:
            best_loss=loss
            best_encoder=encoder.state_dict()  
            best_decoder=decoder.state_dict()
            if save_path:
                torch.save({
                    'epoch':epoch+epoch_trained,
                    'best_encoder_state_dict':best_encoder,
                    'best_decoder_state_dict':best_decoder,
                    'encoder_optimizer_state_dict':encoder_optimizer.state_dict(),
                    'decoder_optimizer_state_dict':decoder_optimizer.state_dict(),
                    'loss':best_loss},
                    save_path)                

    showPlot(plot_losses)
    
def showPlot(points):
    plt.figure()
    fig, ax = plt.subplots()
    # this locator puts ticks at regular intervals
    loc = ticker.MultipleLocator(base=0.2)
    ax.yaxis.set_major_locator(loc)
    plt.plot(points)    
    
def evaluate(encoder, decoder, sentence, input_lang, output_lang,device):
    with torch.no_grad():
        input_tensor = tensorFromSentence(input_lang, sentence,device)

        encoder_outputs, encoder_hidden = encoder(input_tensor)
        decoder_outputs, decoder_hidden, decoder_attn = decoder(encoder_outputs, encoder_hidden)

        _, topi = decoder_outputs.topk(1)
        decoded_ids = topi.squeeze()

        decoded_words = []
        for idx in decoded_ids:
            if idx.item() == EOS_token:
                decoded_words.append('<EOS>')
                break
            decoded_words.append(output_lang.index2word[idx.item()])
    return decoded_words, decoder_attn

def seq2seq_evaluateRandomly(encoder, decoder, pairs,input_lang, output_lang,device,n=10):
    for i in range(n):
        pair = random.choice(pairs)
        print('>', pair[0])
        print('=', pair[1])
        output_words, _ = evaluate(encoder, decoder, pair[0], input_lang, output_lang,device)
        output_sentence = ' '.join(output_words)
        print('<', output_sentence)
        print('')    
        
def showAttention(input_sentence, output_words, attentions):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    cax = ax.matshow(attentions.cpu().numpy(), cmap='bone')
    fig.colorbar(cax)

    # Set up axes
    ax.set_xticklabels([''] + input_sentence.split(' ') +
                       ['<EOS>'], rotation=90)
    ax.set_yticklabels([''] + output_words)

    # Show label at every tick
    ax.xaxis.set_major_locator(ticker.MultipleLocator(1))
    ax.yaxis.set_major_locator(ticker.MultipleLocator(1))
    # print('-'*50)
    plt.show()

def evaluateAndShowAttention(encoder, decoder, input_sentence, input_lang, output_lang,device):
    output_words, attentions = evaluate(encoder, decoder, input_sentence, input_lang, output_lang,device)
    print('input =', input_sentence)
    print('output =', ' '.join(output_words))    
    showAttention(input_sentence, output_words, attentions[0, :len(output_words), :])    
    # print('-'*50)    

if __name__=="__main__":
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    
    root=r'I:\data\NLP_dataset' 
    MAX_LENGTH = 10
    eng_prefixes = (
        "i am ", "i m ",
        "he is", "he s ",
        "she is", "she s ",
        "you are", "you re ",
        "we are", "we re ",
        "they are", "they re "
    )
    
    hidden_size = 128
    batch_size = 32    
    lang1, lang2='eng', 'cmn'
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    input_lang, output_lang,pairs, train_dataloader = get_dataloader4seq2seq(batch_size,root,lang1, lang2,eng_prefixes,MAX_LENGTH,device,reverse=True,cmn=True)    
    
    # root=r'I:\data\NLP_dataset\eng-fra' # \eng-fra.txt'
    # # eng_fra_root=r'I:\data\NLP_dataset'
    
    # MAX_LENGTH = 10
    
    # eng_prefixes = (
    #     "i am ", "i m ",
    #     "he is", "he s ",
    #     "she is", "she s ",
    #     "you are", "you re ",
    #     "we are", "we re ",
    #     "they are", "they re "
    # )

    # hidden_size = 128
    # batch_size = 32    
    # lang1, lang2='eng', 'fra'
    # input_lang, output_lang,pairs, train_dataloader = get_dataloader4seq2seq(batch_size,root,lang1, lang2,eng_prefixes,MAX_LENGTH,device)
    
    
    encoder = EncoderRNN(input_lang.n_words, hidden_size).to(device)
    decoder = AttnDecoderRNN(hidden_size, output_lang.n_words,MAX_LENGTH=MAX_LENGTH,device=device).to(device)
    
    save_path=r'I:\model_ckpts\seq2seq\seq2seq_ch2en.pth'
    seq2seq_train(train_dataloader, encoder, decoder, 80, print_every=5, plot_every=5,save_path=save_path)
    
    checkpoint=torch.load(save_path)
    encoder.load_state_dict(checkpoint['best_encoder_state_dict'])    
    encoder.to(device)    
    decoder.load_state_dict(checkpoint['best_decoder_state_dict'])    
    decoder.to(device)            
    
    encoder.eval()
    decoder.eval()
    seq2seq_evaluateRandomly(encoder, decoder,pairs,input_lang, output_lang,device)    
    
    # evaluateAndShowAttention('il n est pas aussi grand que son pere')
    # evaluateAndShowAttention('je suis trop fatigue pour conduire')
    # evaluateAndShowAttention('je suis desole si c est une question idiote')
    # evaluateAndShowAttention('je suis reellement fiere de vous')  
    
    sentence='我 们 非 常 需 要 食 物'
    input_sentence=' '.join(f'{sentence}')
    evaluateAndShowAttention(encoder, decoder, input_sentence, input_lang, output_lang,device)
    
